# taken from SAMOSA DPM Doc 2.5.2

# S-3 Ku Band SAR Re-tracker Configuration File
CONST_A = 6378137  # Equatorial radius [m]
CONST_B = 6356752.3142  # Polar radius [m]
# CONST_B = 6.356752314245179e+06  # Polar radius [m]
CONST_C = 299792458  # speed of light [m/s]
CONST_F = 0.003352810664747  # flattening coefficient f from WGS84
